import sys
sys.path.append('../')

import os
import yaml
import shutil
import pickle
import os.path as osp
from tqdm import tqdm
from itertools import combinations

import numpy as np
import torch
from torch_geometric.data import Data, InMemoryDataset

from ..utils import get_random_idx_split, download_url, extract_zip, decide_download


class ActsTrack(InMemoryDataset):

    def __init__(self, root, tesla, data_config, transform, device='cpu', seed=0):
        self.url_raw = 'https://zenodo.org/record/7265547/files/actstrack_raw_2T.zip'
        self.url_processed = 'https://zenodo.org/record/7265547/files/actstrack_processed_2T.zip'
        self.tesla = tesla
        self.split = data_config['split']
        self.sample_tracks = data_config['sample_tracks']
        self.pos_features = data_config['pos_features']
        self.other_features = data_config['other_features']
        self.seed = seed

        self.im_thres = data_config['im_thres']  # invariant mass threshold

        super().__init__(root, transform=transform)
        self.data, self.slices, self.idx_split = torch.load(self.processed_paths[0])
        self.data = self.data.to(device)
        self.x_dim = self.data.x.shape[1]
        self.pos_dim = self.data.pos.shape[1]
        self.feature_type = data_config['feature_type']
        self.signal_class = 1
        self.dataset_name = f'actstrack_{self.tesla}'

        if self.feature_type == 'only_pos':
            node_scalar_feat = self.pos_dim
        elif self.feature_type == 'only_x':
            node_scalar_feat = self.x_dim
        elif self.feature_type == 'only_ones':
            node_scalar_feat = 1
        else:
            assert self.feature_type == 'both_x_pos'
            node_scalar_feat = self.x_dim + self.pos_dim

        self.feat_info = {'node_categorical_feat': [], 'node_scalar_feat': node_scalar_feat}

    @property
    def raw_dir(self) -> str:
        return osp.join(self.root, f'raw_{self.tesla}')

    @property
    def processed_dir(self) -> str:
        return osp.join(self.root, f'processed_{self.tesla}')

    @property
    def raw_file_names(self):
        return [f'bkg_events_{self.tesla}.pkl', f'signal_events_{self.tesla}.pkl']

    @property
    def processed_file_names(self):
        return [f'data_{self.tesla}.pt']

    def download(self):
        if self.tesla != '2T':
            raise NotImplementedError('Please download datasets with other magnetic field strength at https://zenodo.org/record/7265547')

        if osp.exists(self.processed_paths[0]):
            return
        if decide_download(self.url_raw, is_raw=True):
            path = download_url(self.url_raw, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
        else:
            if decide_download(self.url_processed, is_raw=False):
                path = download_url(self.url_processed, self.root)
                extract_zip(path, self.root)
                os.unlink(path)
            else:
                print('Stop downloading.')
                shutil.rmtree(self.root)
                exit(-1)

    def process(self):
        if decide_download(self.url_processed, is_raw=False):
            path = download_url(self.url_processed, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
            return

        signal_events = pickle.load(open(self.raw_dir + f'/signal_events_{self.tesla}.pkl', 'rb'))
        signal_data_list = self.build_data(signal_events, 'signal')

        bkg_events = pickle.load(open(self.raw_dir + f'/bkg_events_{self.tesla}.pkl', 'rb'))
        bkg_data_list = self.build_data(bkg_events, 'bkg')

        data_list = signal_data_list + bkg_data_list
        idx_split = get_random_idx_split(len(data_list), self.split, self.seed)

        data, slices = self.collate(data_list)
        torch.save((data, slices, idx_split), self.processed_paths[0])

    def build_data(self, events, event_type):
        data_list = []
        cnt = 0
        pbar = tqdm(events)
        for initial, _, hits in pbar:
            muons = initial[(initial['particle_type'] == 13) | (initial['particle_type'] == -13)]
            electrons = initial[(initial['particle_type'] == 11) | (initial['particle_type'] == -11)]
            if len(hits) == 0 or len(initial) == 0:
                continue

            hits['node_label'] = 0
            y = torch.tensor(0).float().view(-1, 1)
            signal_im = -1
            signal_particles = []
            if event_type == 'signal':
                if len(muons) < 2 and len(electrons) < 2:
                    continue

                signal_electrons = self.get_signal_particles(electrons, self.im_thres)
                signal_muons = self.get_signal_particles(muons, self.im_thres)
                signal_info = np.array(signal_electrons + signal_muons)
                if signal_info.shape[0] != 1:
                    continue

                signal_particles = list(signal_info[:, :2].reshape(-1))
                signal_im = signal_info[:, 2].item()
                assert len(signal_particles) == 2

                hits.loc[hits['particle_id'].isin(signal_particles), 'node_label'] = 1
                y = torch.tensor(1).float().view(-1, 1)

                if hits['node_label'].sum() == 0:  # no signal hits in tracks, even though there are in the initial position
                    continue

            # sampling tracks
            if self.sample_tracks:
                n_ptcl_to_sample = self.sample_tracks - len(signal_particles)
                to_sample = np.random.choice(hits['particle_id'].unique(), n_ptcl_to_sample)
                ptcl_to_use = list(to_sample) + list(signal_particles)
                hits = hits[hits['particle_id'].isin(ptcl_to_use)].reset_index(drop=True)

            hits['node_id'] = range(len(hits))
            pos = torch.tensor(hits[self.pos_features].to_numpy()).float()
            x = torch.tensor(hits[self.other_features].to_numpy()).float()
            node_label = torch.tensor(hits['node_label'].to_numpy()).float().view(-1)
            node_dir = torch.tensor(hits[['tpx', 'tpy', 'tpz']].to_numpy()).float()
            track_ids = torch.full((len(hits),), -1)  # indices which track the node belongs to

            num_tracks = 0
            all_ptcls = hits['particle_id'].unique()
            for ptcl in all_ptcls:
                track = hits[hits['particle_id'] == ptcl]

                if ptcl in signal_particles:
                    assert np.all(track['node_label'] == 1)
                else:
                    assert np.all(track['node_label'] == 0)
                track_ids[track['node_id'].to_numpy()] = num_tracks
                num_tracks += 1
            assert -1 not in track_ids

            data_list.append(Data(x=x, pos=pos, y=y, node_label=node_label,
                                  node_dir=node_dir, num_tracks=num_tracks, track_ids=track_ids, signal_im=signal_im))
            cnt += 1
            pbar.set_description(f'[INFO] Processed {cnt} events')

        return data_list

    @staticmethod
    def invariant_mass(m, px1, py1, pz1, px2, py2, pz2):
        first_term = m**2
        second_term = np.sqrt(m**2 + px1**2 + py1**2 + pz1**2) * np.sqrt(m**2 + px2**2 + py2**2 + pz2**2)
        third_term = px1*px2 + py1*py2 + pz1*pz2
        return np.sqrt(2 * (first_term + second_term - third_term))

    @staticmethod
    def get_signal_particles(particles, thres):
        if len(particles) < 2:
            return []

        res = []
        all_particle_pairs = combinations(range(len(particles)), 2)
        for i, j in all_particle_pairs:
            first_particle, second_particle = particles.iloc[i], particles.iloc[j]
            if first_particle['q'] * second_particle['q'] > 0:
                continue

            # if np.linalg.norm(first_particle[['vx', 'vy', 'vz']] - second_particle[['vx', 'vy', 'vz']], ord=2) < 0.1:
            im = ActsTrack.invariant_mass(first_particle['m'], first_particle['px'], first_particle['py'], first_particle['pz'], second_particle['px'], second_particle['py'], second_particle['pz'])
            if abs(im - 91.1876) < thres:
                res.append([first_particle['particle_id'], second_particle['particle_id'], im])
        return res


if __name__ == '__main__':
    data_config = yaml.safe_load(open('../configs/egnn_actstrack.yml'))['data']
    # for tesla in tqdm(['2T', '4T', '6T', '8T', '10T', '12T', '14T', '16T', '18T', '20T']):
    tesla = '2T'
    dataset = ActsTrack(root='../../data/actstrack', tesla=tesla, data_config=data_config, seed=42)



# https://github.com/pyg-team/pytorch_geometric/blob/master/examples/point_transformer_classification.py

from torch import Tensor
from typing import Callable, Optional, Tuple, Union
from torch_geometric.typing import Adj, OptTensor, PairTensor

import torch
import torch.nn.functional as F
from torch.nn import Linear as Lin

from torch_geometric.nn.inits import reset
from torch_sparse import SparseTensor, set_diag
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.nn.dense.linear import Linear
from torch_geometric.utils import add_self_loops, remove_self_loops, softmax

from ..utils import FeatEncoder


class PointTransformer(torch.nn.Module):
    def __init__(self, x_dim, pos_dim, model_config, feat_info, n_categorical_feat_to_use=-1, n_scalar_feat_to_use=-1, **kwargs):
        super().__init__()

        hidden_size = model_config['hidden_size']
        self.n_layers = model_config['n_layers']

        self.x_dim = x_dim
        self.pos_dim = pos_dim
        self.dropout_p = model_config['dropout_p']
        self.raw_pos_dim = kwargs['aux_info']['raw_pos_dim']

        self.node_encoder = FeatEncoder(hidden_size, feat_info['node_categorical_feat'], feat_info['node_scalar_feat'], n_categorical_feat_to_use, n_scalar_feat_to_use)
        self.edge_encoder = FeatEncoder(hidden_size, feat_info['edge_categorical_feat'], feat_info['edge_scalar_feat'])

        self.convs = torch.nn.ModuleList()
        for _ in range(self.n_layers):
            self.convs.append(TransformerBlock(hidden_size, hidden_size, pos_dim=self.raw_pos_dim))

    def forward(self, x, pos, edge_attr, edge_index, batch, edge_attn=None):
        if self.x_dim == 0 and self.pos_dim != 0:
            feats = pos
        elif self.x_dim != 0 and self.pos_dim == 0:
            feats = x
        elif self.x_dim == 0 and self.pos_dim == 0:
            feats = torch.ones(x.shape[0], 1, device=x.device)
        else:
            feats = torch.cat([x, pos], dim=1)

        x = self.node_encoder(feats)
        edge_attr = self.edge_encoder(edge_attr)

        for i in range(self.n_layers):
            identity = x
            x = self.convs[i](x, pos, edge_index, edge_attr=edge_attr, edge_attn=edge_attn)
            x = x + identity
            x = F.dropout(x, self.dropout_p, training=self.training)
        return x


class TransformerBlock(torch.nn.Module):
    def __init__(self, in_channels, out_channels, pos_dim):
        super().__init__()
        self.lin_in = Lin(in_channels, in_channels)
        self.lin_out = Lin(out_channels, out_channels)
        self.transformer = PointTransformerConv(in_channels, out_channels, pos_dim=pos_dim)

    def forward(self, x, pos, edge_index, edge_attr=None, edge_attn=None):
        x = self.lin_in(x).relu()
        x = self.transformer(x, pos, edge_index, edge_attr=edge_attr, edge_attn=edge_attn)
        x = self.lin_out(x).relu()
        return x


class PointTransformerConv(MessagePassing):
    r"""The Point Transformer layer from the `"Point Transformer"
    <https://arxiv.org/abs/2012.09164>`_ paper

    .. math::
        \mathbf{x}^{\prime}_i =  \sum_{j \in
        \mathcal{N}(i) \cup \{ i \}} \alpha_{i,j} \left(\mathbf{W}_3
        \mathbf{x}_j + \delta_{ij} \right),

    where the attention coefficients :math:`\alpha_{i,j}` and
    positional embedding :math:`\delta_{ij}` are computed as

    .. math::
        \alpha_{i,j}= \textrm{softmax} \left( \gamma_\mathbf{\Theta}
        (\mathbf{W}_1 \mathbf{x}_i - \mathbf{W}_2 \mathbf{x}_j +
        \delta_{i,j}) \right)

    and

    .. math::
        \delta_{i,j}= h_{\mathbf{\Theta}}(\mathbf{p}_i - \mathbf{p}_j),

    with :math:`\gamma_\mathbf{\Theta}` and :math:`h_\mathbf{\Theta}`
    denoting neural networks, *i.e.* MLPs, and
    :math:`\mathbf{P} \in \mathbb{R}^{N \times D}` defines the position of
    each point.

    Args:
        in_channels (int or tuple): Size of each input sample, or :obj:`-1` to
            derive the size from the first input(s) to the forward method.
            A tuple corresponds to the sizes of source and target
            dimensionalities.
        out_channels (int): Size of each output sample.
        pos_nn : (torch.nn.Module, optional): A neural network
            :math:`h_\mathbf{\Theta}` which maps relative spatial coordinates
            :obj:`pos_j - pos_i` of shape :obj:`[-1, 3]` to shape
            :obj:`[-1, out_channels]`.
            Will default to a :class:`torch.nn.Linear` transformation if not
            further specified. (default: :obj:`None`)
        attn_nn : (torch.nn.Module, optional): A neural network
            :math:`\gamma_\mathbf{\Theta}` which maps transformed
            node features of shape :obj:`[-1, out_channels]`
            to shape :obj:`[-1, out_channels]`. (default: :obj:`None`)
        add_self_loops (bool, optional) : If set to :obj:`False`, will not add
            self-loops to the input graph. (default: :obj:`True`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.

    Shapes:
        - **input:**
          node features :math:`(|\mathcal{V}|, F_{in})` or
          :math:`((|\mathcal{V_s}|, F_{s}), (|\mathcal{V_t}|, F_{t}))`
          if bipartite,
          positions :math:`(|\mathcal{V}|, 3)` or
          :math:`((|\mathcal{V_s}|, 3), (|\mathcal{V_t}|, 3))` if bipartite,
          edge indices :math:`(2, |\mathcal{E}|)`
        - **output:** node features :math:`(|\mathcal{V}|, F_{out})` or
          :math:`(|\mathcal{V}_t|, F_{out})` if bipartite
    """
    def __init__(self, in_channels: Union[int, Tuple[int, int]],
                 out_channels: int, pos_nn: Optional[Callable] = None,
                 attn_nn: Optional[Callable] = None,
                 add_self_loops: bool = False, pos_dim=3, **kwargs):
        kwargs.setdefault('aggr', 'add')  # https://github.com/pyg-team/pytorch_geometric/pull/5332
        super().__init__(**kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.add_self_loops = add_self_loops

        if isinstance(in_channels, int):
            in_channels = (in_channels, in_channels)

        self.pos_nn = pos_nn
        if self.pos_nn is None:
            self.pos_nn = Linear(pos_dim, out_channels)

        self.attn_nn = attn_nn
        self.lin = Linear(in_channels[0], out_channels, bias=False)
        self.lin_src = Linear(in_channels[0], out_channels, bias=False)
        self.lin_dst = Linear(in_channels[1], out_channels, bias=False)

        self.reset_parameters()

    def reset_parameters(self):
        reset(self.pos_nn)
        if self.attn_nn is not None:
            reset(self.attn_nn)
        self.lin.reset_parameters()
        self.lin_src.reset_parameters()
        self.lin_dst.reset_parameters()

    def forward(
        self,
        x: Union[Tensor, PairTensor],
        pos: Union[Tensor, PairTensor],
        edge_index: Adj, edge_attr=None, edge_attn=None
    ) -> Tensor:
        """"""
        if isinstance(x, Tensor):
            alpha = (self.lin_src(x), self.lin_dst(x))
            x: PairTensor = (self.lin(x), x)
        else:
            alpha = (self.lin_src(x[0]), self.lin_dst(x[1]))
            x = (self.lin(x[0]), x[1])

        if isinstance(pos, Tensor):
            pos: PairTensor = (pos, pos)

        if self.add_self_loops:
            if isinstance(edge_index, Tensor):
                edge_index, _ = remove_self_loops(edge_index)
                edge_index, _ = add_self_loops(
                    edge_index, num_nodes=min(pos[0].size(0), pos[1].size(0)))
            elif isinstance(edge_index, SparseTensor):
                edge_index = set_diag(edge_index)

        # propagate_type: (x: PairTensor, pos: PairTensor, alpha: PairTensor)
        out = self.propagate(edge_index, x=x, pos=pos, alpha=alpha, size=None, edge_attr=edge_attr, edge_attn=edge_attn)
        return out

    def message(self, x_j: Tensor, pos_i: Tensor, pos_j: Tensor,
                alpha_i: Tensor, alpha_j: Tensor, index: Tensor,
                ptr: OptTensor, size_i: Optional[int], edge_attr=None, edge_attn=None) -> Tensor:

        delta = self.pos_nn(pos_i - pos_j)
        alpha = alpha_i - alpha_j + delta
        if self.attn_nn is not None:
            alpha = self.attn_nn(alpha)
        alpha = softmax(alpha, index, ptr, size_i)

        if edge_attr is not None:
            msg = alpha * (x_j + delta + edge_attr)
        else:
            msg = alpha * (x_j + delta)

        if edge_attn is not None:
            msg = msg * edge_attn
        return msg

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels})')



from pathlib import Path
import pandas as pd

def calculate_df(df):
    df[['exp_method', 'clf_method']] = df.index.str.split('_', n=1).tolist()

    # 修正clf_method列，因为它包含了额外的数字
    df[['clf_method', 'seed']] = df['clf_method'].str.rsplit('_', n=1).tolist()

    # 计算每一组的平均值和标准差
    result = df.groupby(['model', 'dataset', 'clf_method', 'exp_method']).agg(
        avg_exp_auc=('exp_auc', 'mean'),
        std_exp_auc=('exp_auc', 'std')
    ).reset_index()
    return result

# 定义一个函数来处理文件和目录
def process_directory(directory_path):
    all_data = []  # 用于存储所有csv文件数据的列表

    # 使用Path对象遍历目录
    path = Path(directory_path)
    
    # 遍历第一级目录，即model_dataset目录
    for model_dataset_dir in path.iterdir():
        if model_dataset_dir.is_dir():
            # 从目录名中提取model和dataset名称
            model, dataset = model_dataset_dir.name.split('_')
            
            # 遍历第二级目录，即explainer_inherent目录
            for explainer_inherent_dir in model_dataset_dir.iterdir():
                if explainer_inherent_dir.is_dir() and explainer_inherent_dir.name == 'explainer_inherent':
                    # 遍历CSV文件
                    for csv_file in explainer_inherent_dir.glob('*.csv'):
                        # 读取CSV文件到DataFrame
                        df = pd.read_csv(csv_file, index_col=0)
                        # 添加model和dataset列
                        df['model'] = model
                        df['dataset'] = dataset
                        # 将DataFrame添加到列表中
                        all_data.append(df)
    
    # 合并所有DataFrame
    final_df = pd.concat(all_data)
    return final_df

# 调用函数并传入顶级目录路径
df = process_directory('./result')
new_df = df[['test_exp_auc', 'model', 'dataset']].rename(columns={'test_exp_auc': 'exp_auc'})
final_df = calculate_df(new_df)

# reindex and multicolumn and combine results
final_df['avg_std'] = final_df.apply(lambda row: f"${row['avg_exp_auc']*100:.2f} \pm {row['std_exp_auc']:.2f}$", axis=1)
pivot_df = final_df.pivot_table(index=['clf_method', 'exp_method'], columns=['dataset', 'model'], values=["avg_std"], aggfunc='first')

# output final table
pd.set_option('display.max_columns', None)
print(pivot_df)
pivot_df.to_csv('./new-post-inherent.csv', index=True)





import sys
sys.path.append('../')

import os
import yaml
import shutil
import os.path as osp
from tqdm import tqdm

import numpy as np
import pandas as pd
import torch
from torch_geometric.data import Data, InMemoryDataset

from ..utils import log, get_random_idx_split, download_url, extract_zip, decide_download


class Tau3Mu(InMemoryDataset):
    def __init__(self, root, data_config, transform, device='cpu', seed=0):
        self.url_raw = 'https://zenodo.org/record/7265547/files/tau3mu_raw.zip'
        self.url_processed = 'https://zenodo.org/record/7265547/files/tau3mu_processed.zip'
        self.split = data_config['split']
        self.other_features = data_config['other_features']
        self.seed = seed

        self.sample_filters = data_config['sample_filters']
        self.hit_filters = data_config['hit_filters']

        super().__init__(root, transform=transform)
        self.data, self.slices, self.idx_split = torch.load(self.processed_paths[0])
        self.data = self.data.to(device)
        self.x_dim = self.data.x.shape[1]
        self.pos_dim = self.data.pos.shape[1]
        self.signal_class = 1
        self.dataset_name = 'tau3mu'
        self.feature_type = data_config['feature_type']

        if self.feature_type == 'only_pos':
            node_scalar_feat = self.pos_dim
        elif self.feature_type == 'only_x':
            node_scalar_feat = self.x_dim
        elif self.feature_type == 'only_ones':
            node_scalar_feat = 1
        else:
            assert self.feature_type == 'both_x_pos'
            node_scalar_feat = self.x_dim + self.pos_dim

        self.feat_info = {'node_categorical_feat': [], 'node_scalar_feat': node_scalar_feat}

    @property
    def raw_file_names(self):
        return ['tau3mu_mixed.pkl']

    @property
    def processed_file_names(self):
        return ['data.pt']

    def download(self):
        if osp.exists(self.processed_paths[0]):
            return
        if decide_download(self.url_raw, is_raw=True):
            path = download_url(self.url_raw, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
        else:
            if decide_download(self.url_processed, is_raw=False):
                path = download_url(self.url_processed, self.root)
                extract_zip(path, self.root)
                os.unlink(path)
            else:
                print('Stop downloading.')
                shutil.rmtree(self.root)
                exit(-1)

    def process(self):
        if decide_download(self.url_processed, is_raw=False):
            path = download_url(self.url_processed, self.root)
            extract_zip(path, self.root)
            os.unlink(path)
            return

        df = pd.read_pickle(self.raw_dir + '/tau3mu_mixed.pkl')

        log('[INFO] Processing entries...')
        data_list = []
        for entry in tqdm(df.itertuples(), total=len(df)):
            entry = self.mask_hits(entry, self.hit_filters, self.sample_filters)
            if entry is None:
                continue
            x = torch.tensor(np.stack([entry[feature] for feature in self.other_features], axis=1)).float()
            pos = self.get_pos(entry)
            y = torch.tensor(entry['y']).float().view(-1, 1)

            if y.item() == 1:
                node_label = torch.tensor(entry['node_label']).float().view(-1)
            else:
                node_label = torch.zeros((x.shape[0])).float()

            data = Data(x=x, pos=pos, y=y, node_label=node_label)
            data_list.append(data)

        idx_split = get_random_idx_split(len(data_list), self.split, self.seed)
        data, slices = self.collate(data_list)

        log('[INFO] Saving data.pt...')
        torch.save((data, slices, idx_split), self.processed_paths[0])

    @staticmethod
    def get_pos(entry):
        x = entry['mu_hit_sim_eta'].reshape(-1, 1)
        y = np.deg2rad(entry['mu_hit_sim_phi'].reshape(-1, 1))
        return torch.tensor(np.concatenate((x, y), axis=1)).float()

    @staticmethod
    def mask_hits(entry, hit_filters, sample_filters):
        mask = np.ones(entry.n_mu_hit, dtype=bool)
        for k, v in hit_filters.items():
            assert isinstance(getattr(entry, k), np.ndarray)
            mask *= eval('entry.' + k + v)

        masked_entry = {'n_mu_hit': mask.sum()}
        for k in entry._fields:
            value = getattr(entry, k)
            if isinstance(value, np.ndarray) and 'gen' not in k and k != 'y' and 'L1' not in k:
                assert value.shape[0] == entry.n_mu_hit
                assert 'mu_hit' in k or k == 'node_label'
                masked_entry[k] = value[mask].reshape(-1)
            else:
                if k != 'n_mu_hit':
                    masked_entry[k] = value

        if masked_entry['y'] == 1 and not eval("masked_entry['node_label'].sum()" + sample_filters['num_hits']):
            return None
        if masked_entry['y'] == 0 and not eval("masked_entry['n_mu_hit']" + sample_filters['num_hits']):
            return None
        return masked_entry


if __name__ == '__main__':
    data_config = yaml.safe_load(open('../configs/tau3mu.yml'))['data']
    dataset = Tau3Mu(root='../../data/tau3mu', data_config=data_config, seed=42)



import yaml
import json
import argparse
from tqdm import tqdm
from pathlib import Path
from sklearn.metrics import roc_auc_score
import scipy.stats
import torch
import torch.nn as nn
from torch.nn import functional as F
from eval import FidelEvaluation, AUCEvaluation
from get_model import Model
from baselines import *
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, load_checkpoint, ExtractorMLP, get_optimizer
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
import numpy as np
import pandas as pd
import warnings
from datetime import datetime
from scipy.stats import spearmanr as SP
warnings.filterwarnings("ignore")

def eval_one_batch(baseline, data, epoch):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradx', 'inter_grad', 'gnnexplainer']):
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP

        loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=False)

        if baseline.name == 'vgib':
            infer_clf_logits = baseline.forward(data)
        # from sklearn.metrics import roc_auc_score
        # roc_auc_score(infer_clf_logits, data.y.cpu())
        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)


def ensemble(config, method_name, model_name, backbone_seed, dataset_name, log_dir, device):
    set_seed(backbone_seed)
    print('The logging directory is', log_dir), print('=' * 80)

    # batch_size = config['optimizer']['batch_size']
    batch_size = 1
    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],  # backbone_config
                method_name, config[method_name],  # method_config
                dataset).to(device)

    extractor = ExtractorMLP(config[model_name]['hidden_size'], config[method_name], config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits

    map_location = torch.device('cpu') if not torch.cuda.is_available() else None
    baseline = eval(name_mapping[method_name])(clf, extractor, criterion, config[method_name]) if method_name in inherent_models + ['pgexplainer'] \
        else eval(name_mapping[method_name])(clf, criterion, config[method_name])

    if method_name in inherent_models:
        assert load_checkpoint(baseline, log_dir, model_name=method_name, seed=backbone_seed, map_location=map_location,
                               backbone_seed=backbone_seed)
    elif method_name == 'pgexplainer':
        assert load_checkpoint(baseline, log_dir, model_name=method_name, seed=backbone_seed, map_location=map_location,
                        backbone_seed=backbone_seed)
    else:
        print(f'The method {method_name} has no extra model.')

    if method_name not in inherent_models:
        erm_dir = log_dir.parent / 'erm'
        assert load_checkpoint(baseline.clf, erm_dir, model_name='erm', seed=backbone_seed, map_location=map_location)
    # establish the model and the metrics

    baseline.start_tracking() if 'grad' in method_name or method_name == 'gnnlrp' else None
    metric_list = [AUCEvaluation()] if batch_size > 64 else None

    pbar, avg_loss_dict = tqdm(loaders['test']), dict()
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    epoch_attn, epoch_label, epoch_sig, epoch_bkg = [], [], [], []
    # sp_list = []
    for idx, data in enumerate(pbar):
        if data.y.item() == 0:
            continue

        eval_dict = {}        # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
        loss_dict, clf_logits, attn = eval_one_batch(baseline, data.to(baseline.device), 1)
        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()
        # print()
        # min_max_sclarer
        attn = (attn - attn.min()) / (attn.max() - attn.min()) if method_name not in inherent_models else attn

        epoch_attn.append(attn), epoch_label.append(ex_labels)
        epoch_sig.append(attn[torch.where(data.node_label == 1)]), epoch_bkg.append(attn[torch.where(data.node_label == 0)])

        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
        eval_dict.update({f'exp_auc': roc_auc_score(ex_labels, attn)})

        # print('The spearman correlation between avg and std:', )
        desc = log_epoch(backbone_seed, 1, 'test', loss_dict, eval_dict, phase_info='dataset')
        pbar.set_description(desc)
        # compute the avg_loss for the epoch desc
        exec('for k, v in loss_dict.items():\n\tavg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)')

    # epoch_dict = {eval_metric.name: eval_metric.eval_epoch(return_att=True) for eval_metric in metric_list} if metric_list else {}
    # assert or  metric_list[0].name == 'exp_auc'
    epoch_attn, epoch_label, epoch_sig, epoch_bkg = torch.cat(epoch_attn), torch.cat(epoch_label), torch.cat(epoch_sig), torch.cat(epoch_bkg)
    epoch_dict = {'exp_auc': roc_auc_score(epoch_label, epoch_attn)}
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    print('-' * 80, '\n', 'epoch_dict:', epoch_dict, '\n', '-' * 80)

    return epoch_attn, epoch_label, epoch_sig, epoch_bkg


def save_multi_bseeds(multi_methods_res, method_name, exp_method, seeds):
    seeds += ['avg', 'std']
    indexes = [method_name+'_'+str(seed) for seed in seeds]
    # from itertools import product
    # indexes = ['_'.join(item) for item in product(methods, seeds)]
    df = pd.DataFrame(multi_methods_res, index=indexes)

    day_dir = Path('../result') / config_name / datetime.now().strftime("%m_%d")
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([exp_method, method_name, 'sensitivity.csv']))
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    return df


def save_multi_result(method_res, all_methods, config_name):

    df = pd.DataFrame(method_res, index=all_methods)

    day_dir = Path('../result') / config_name / 'data_insights'
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join(all_methods) + '_insights.csv')
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    now_df = pd.read_csv(csv_dir, index_col=0)
    return now_df


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack')
    parser.add_argument('--method', type=str, help='method used')
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--seeds', type=int, nargs="+", help='random seed for data insights pipeline', default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    # parser.add_argument('--mseed', type=int, help='random seed for explainer', default=0)
    parser.add_argument('--use_csv', action="store_true")

    args = parser.parse_args()
    # sub_metric = 'avg_loss'
    print(args)

    dataset_name, model_name, cuda_id, note = args.dataset, args.backbone, args.cuda, args.note
    config_name = '_'.join([model_name, dataset_name])
    device = torch.device(f'cuda:{cuda_id}' if cuda_id >= 0 and torch.cuda.is_available() else 'cpu')
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('../configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    all_methods = post_hoc_attribution + post_hoc_explainers + inherent_models if args.method == 'all' else [args.method]
    assert args.use_csv
    method_res = []
    for method_name in all_methods:
        method_dict = {'single_attn_auc': []}
        if config[method_name].get(model_name, False):
            config[method_name].update(config[method_name][model_name])
        print('=' * 80)
        print(f'Config for {method_name}: ', json.dumps(config[method_name], indent=4))

        model_dir = Path('../log') / config_name / method_name
        multi_attn, multi_sig_attn, multi_bkg_attn = None, None, None

        for all_seed in args.seeds:
            # we directly load attns from csv file
            mseed = [0] if method_name in post_hoc_attribution else [all_seed] if method_name in inherent_models else \
                list(range(10))
            save_dir = model_dir / f"bs{all_seed}_ms{mseed}_attns.csv"
            seed_df = pd.read_csv(save_dir, index_col=0)
            epoch_label = torch.tensor(seed_df['node_labels'])
            if 'attn' in seed_df:  # this means the method has no need to vary seeds for fixed backbone_seed
                seed_attn = torch.tensor(seed_df['attn']).unsqueeze(1)
            else:  # for post-hoc explainers, there are many seed res and only use the same seed with backbone_seed
                seed_attn = torch.tensor(seed_df[str(all_seed)]).unsqueeze(1)
            method_dict['single_attn_auc'].append(roc_auc_score(epoch_label, seed_attn))
            seed_sig_attn = torch.tensor(seed_attn[seed_df['node_labels'] == 1]).unsqueeze(1)
            seed_bkg_attn = torch.tensor(seed_attn[seed_df['node_labels'] == 0]).unsqueeze(1)
            multi_attn = seed_attn if multi_attn is None else torch.cat([multi_attn, seed_attn], dim=1)
            multi_sig_attn = seed_sig_attn if multi_sig_attn is None else torch.cat([multi_sig_attn, seed_sig_attn], dim=1)
            multi_bkg_attn = seed_bkg_attn if multi_bkg_attn is None else torch.cat([multi_bkg_attn, seed_bkg_attn], dim=1)

        avg_attn, std_attn = torch.mean(multi_attn, dim=1), torch.std(multi_attn, dim=1)
        avg_sig_attn, std_sig_attn = torch.mean(multi_sig_attn, dim=1), torch.std(multi_sig_attn, dim=1)
        avg_bkg_attn, std_bkg_attn = torch.mean(multi_bkg_attn, dim=1), torch.std(multi_bkg_attn, dim=1)

        exp_auc_avg, exp_auc_std = roc_auc_score(epoch_label, avg_attn), roc_auc_score(epoch_label, 1 - std_attn)
        all_spearman, sig_spear, bkg_spear = SP(avg_attn, std_attn)[0], SP(avg_sig_attn, std_sig_attn)[0], SP(avg_bkg_attn, std_bkg_attn)[0]
        method_dict.update({'avg_attn_auc': roc_auc_score(epoch_label, avg_attn), 'std_attn_auc': roc_auc_score(epoch_label, 1-std_attn),
                        'all_attn_avg': avg_attn.mean().item(), 'sig_attn_avg': avg_sig_attn.mean().item(), 'bkg_attn_avg': avg_bkg_attn.mean().item(),
                       'all_attn_std': std_attn.mean().item(), 'sig_attn_std': std_sig_attn.mean().item(), 'bkg_attn_std': std_bkg_attn.mean().item()})
        method_dict.update({'single_attn_auc_avg': np.mean(method_dict['single_attn_auc']),
                            'single_attn_auc_std': np.var(method_dict['single_attn_auc'])})
        method_res += [method_dict]
        print(json.dumps(method_dict, indent=4))

    save_multi_result(method_res, all_methods, config_name)





import torch
import torch.nn as nn
import torch.nn.functional as F
from torch_geometric.utils import to_dense_adj
from ..base import BaseRandom


class VGIB(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'vgib'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device
        self.mse_loss = nn.MSELoss()
        self.mi_loss_coef = config['mi_loss_coef']
        self.reg_loss_coef = config['reg_loss_coef']
        self.noise_loss_coef = config['noise_loss_coef']
        self.sample = 'both'

    def __loss__(self, org_logits, noisy_logits, clf_labels, mi_loss, reg_loss):
        pred_loss = self.criterion(org_logits, clf_labels.float())
        noise_loss = self.criterion(noisy_logits, clf_labels.float())

        # r = self.get_r(self.epoch)
        # info_loss = (attn * torch.log(attn/r + 1e-6) + (1 - attn) * torch.log((1 - attn)/(1 - r + 1e-6) + 1e-6)).mean()

        mi_loss = self.mi_loss_coef * mi_loss
        reg_loss = self.reg_loss_coef * reg_loss
        noise_loss = self.noise_loss_coef * noise_loss

        loss = pred_loss + noise_loss + mi_loss + reg_loss
        loss_dict = {'pred': pred_loss.item(), 'noise': noise_loss.item(), 'mi': mi_loss.item(), 'reg': reg_loss.item()}
        return loss, loss_dict

    def forward(self, data, edge_attn=None):
        epsilon = 0.0000001
        emb, edge_index = self.clf.get_emb(data, edge_attn=edge_attn)
        original_clf_logits = self.clf.get_pred_from_emb(emb, data.batch)
        return original_clf_logits

    def forward_pass(self, data, epoch, do_sampling, **kwargs):
        epsilon = 0.0000001
        emb, edge_index = self.clf.get_emb(data)
        original_clf_logits = self.clf.get_pred_from_emb(emb, data.batch)


        attn_log_logits = self.extractor(emb, batch=data.batch, pool_out_lig=None)
        attn = attn_log_logits.sigmoid()
        # attn = F.softmax(attn_log_logits, dim=1)
        # edge_attn = self.node_attn_to_edge_attn(attn, data.edge_index) if att_type == 'node' else attn if att_type == 'edge' else None

        #this part is used to add noise
        # node_feature = emb
        static_node_feature = emb.clone().detach()
        node_feature_std, node_feature_mean = torch.std_mean(static_node_feature, dim=0)

        #this part is used to generate assignment matrix
        # abstract_features_1 = torch.tanh(self.fully_connected_1(node_feature))

        assignment = torch.cat([attn, 1-attn], dim=1)
        gumble_assign = F.gumbel_softmax(assignment)

        #This is the graph embedding
        # graph_feature = torch.sum(node_feature, dim=0, keepdim=True)

        #add noise to the node representation
        node_feature_mean = node_feature_mean.repeat(emb.shape[0], 1)

        #noisy_graph_representation
        lambda_pos = gumble_assign[:, 0].unsqueeze(dim=1)
        if self.sample == 'both':
            lambda_neg = gumble_assign[:, 1].unsqueeze(dim=1)
        else:
            assert self.sample == 'single'
            lambda_neg = (1-lambda_pos)


        #this is subgraph embedding
        # subgraph_representation = torch.sum(lambda_pos * node_feature, dim=0, keepdim=True)

        # the noise confer to the distribution of
        noisy_node_feature_mean = lambda_pos * emb + lambda_neg * node_feature_mean
        noisy_node_feature_std = lambda_neg * node_feature_std
        noisy_emb = noisy_node_feature_mean + torch.rand_like(noisy_node_feature_mean) * noisy_node_feature_std
        # with torch.no_grad():
        noisy_clf_logits = self.clf.get_pred_from_emb(noisy_emb, data.batch)

        KL_tensor = 0.5 * ((noisy_node_feature_std ** 2) / (node_feature_std+epsilon) ** 2) + \
                    torch.sum(((noisy_node_feature_mean - node_feature_mean)/(node_feature_std + epsilon))**2, dim=0) #+\
        #            torch.log(node_feature_std / (noisy_node_feature_std + epsilon) + epsilon)
        mi_loss = torch.mean(KL_tensor)

        EYE = torch.ones(2, device=KL_tensor.device)
        # Pos_mask = torch.FloatTensor([1, 0])

        Adj = to_dense_adj(edge_index)[0]
        Adj.requires_grad = False
        try:
            new_adj = torch.mm(torch.t(assignment), Adj)
            new_adj = torch.mm(new_adj, assignment)
            normalize_new_adj = F.normalize(new_adj, p=1, dim=1)

            norm_diag = torch.diag(normalize_new_adj)
            reg_loss = self.mse_loss(norm_diag, EYE)
        except:
            print(f"There is some graph in which some node has no edges. ({assignment.shape[0]}-{Adj.shape[0]})")
            reg_loss = torch.tensor(0, device=original_clf_logits.device)

        loss, loss_dict = self.__loss__(original_clf_logits, noisy_clf_logits, data.y, mi_loss, reg_loss)
        #cal preserve rate

        attn = self.node_attn_to_edge_attn(attn, edge_index) if hasattr(data, 'edge_label') else attn

        return loss, loss_dict, noisy_clf_logits, attn.reshape(-1)

    def get_r(self, current_epoch):
        r = self.init_r - current_epoch // self.decay_interval * self.decay_r
        if r < self.final_r:
            r = self.final_r
        return r

    def sampling(self, attn_log_logits, do_sampling):
        if do_sampling:
            random_noise = torch.empty_like(attn_log_logits).uniform_(1e-10, 1 - 1e-10)
            random_noise = torch.log(random_noise) - torch.log(1.0 - random_noise)
            attn_bern = ((attn_log_logits + random_noise) / self.temperature).sigmoid()
        else:
            attn_bern = (attn_log_logits).sigmoid()
        return attn_bern

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        # edge_attn = 1 - (1 - src_attn) * (1 - dst_attn)
        edge_attn = src_attn * dst_attn
        # edge_attn = (src_attn + dst_attn) / 2
        return edge_attn
